/* Copyright (C) Eelco Vriezekolk, Universiteit Twente, Agentschap Telecom.
 * See LICENSE.md
 */

/* global
 Component, ComponentIterator, DefaultThreats, H, LS, NodeCluster, NodeCluster, NodeClusterIterator, NodeClusterIterator, Preferences, Service, ServiceIterator, Threat, Threat, ThreatAssessment, ThreatIterator, ToolGroup, _, bugreport, exportProject, isSameString, loadFromString, newRasterConfirm, nextUnusedIndex, nid2id, prettyDate, rasterAlert, startAutoSave, switchToProject, transactionCompleted, trimwhitespace, urlEncode
*/

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Project: object representing a project
 *
 * Class variables (those prefixed with underscore should not be accessed from outside)
 *	cid: integer ID of the currently active project
 *	_all: array of all projects, indexed on project ID
 *	get(i): returns the object with id 'i'.
 *	withTitle(str): returns the id of the local project with title 'str', or 'null' otherwise.
 *	firstProject(): returns the an existing local Project object, or null otherwise.
 *  merge(target,source): merge source project into target project, deleting source project.
 *	updateStubs(blocking): retrieve list of projects from server.
 *	retrieve(i): download the project with id 'i' from the server.
 *	getProps(s): retrieve properties of remote project with name 's'.
 * Instance properties:
 *	id: (integer) unique ID of the project
 *	title: (string) short name of the project
 *  group: (string) group to which the project belongs (default group is ToolGroup)
 *		group is only meaningful when the project is shared or a stub.
 *		Webserver-based access controls may apply per group.
 *  shared: (boolean) true iff the project is shared, and stored on the server.
 *	description: free-form description (max 100 chars).
 *	services[]: list of service IDs of all services belonging to this project. Empty for stubs.
 *	threats[]: list of all checklist threats belonging to this project. Empty for stubs.
 *	labels[]: list of all label strings.
 *	creator: (string) creator, as returned by the server. Only used for stubs.
 *	date: (string) date, as returned by the server. Only used for stubs.
 *	stub: (boolean) true iff this is a stub.
 * Methods:
 *	destroy(): destructor for this object
 *	totalnodes(): returns the count of all nodes within all services of this project.
 *	settitle(t): change the title to 't' (50 chars max). The actual title may receive
 *		a numerical suffix to make it unique.
 *	setshared(b): set sharing status to 'b' (boolean).
 *	setdescription(s): change description to 's'.
 *	setdate(d): change date to 'd'.
 *	autosettitle: choose a unique standard name.
 *		Either settitle() or autosettitle() must be called.
 *	addservice(id,title): add an existing Service object to this project.
 *	removeservice(id): remove and destroy a Service object from this project.
 *  addthreat(id): add a checklist threat to the project.
 *	removethreat(id): remove a checklist threat from the project.
 *	unload(): remove all DOM elements for this project.
 *	load(): create and set all DOM elements for this project.
 *	adddefaultthreats: add the predefined checklist threats to this project.
 *	_stringify: create a JSON text string representing this object's data.
 *	exportstring: return a line of text for insertion when saving this file.
 *	store(): store the object into localStorage.
 *	storeOnServer(): save this project onto the server.
 *  storeIfNotPresent():
 *	deleteFromServer(): remove this project from the server.
 *	getDate(): retrieve last saved date of this project.
 *  refreshIfNecessary():
 *  dorefresh():
 */
var Project = function(id,asstub) {
	if (id!=null && Project._all[id]!=null) {
		bugreport("Project with id "+id+" already exists","Project.constructor");
	}
	this.id = (id==null ? nextUnusedIndex(Project._all) : id);
	this.title = "?";
	this.group = ToolGroup;
	this.description = "";
	this.services = [];
	this.threats = [];
	this.labels = Project.defaultlabels;
	this.creator = "";
	this.date = "";








	this.stub = false;

	this.store();
	Project._all[this.id]=this;
};
Project.get = function(id) { return Project._all[id]; };
Project.cid = 0;
Project._all = [];
Project.defaultlabels = [_("Red"), _("Orange"), _("Yellow"), _("Green"), _("Blue"), _("Pink"), _("Purple"), _("Grey")];
Project.colors = ["none","red","orange","yellow","green","blue","pink","purple","grey"];

// Check wether there is a project with name 'str' and group ToolGroup
Project.withTitle = function(str) {
	var found=false;
	for (var i=0; !found && i<Project._all.length; i++) {
		var p = Project._all[i];
		if (p==null || p.stub) continue;

		found=(isSameString(p.title,str)  && (!p.shared || p.group==ToolGroup));
	}
	return (found ? i-1 : null);
};

// Retrieve first project in ToolGroup
Project.firstProject = function() {
	for (var i=0; i<Project._all.length; i++) {
		var p = Project._all[i];
		if (p==null || p.stub) continue;
		if (!p.shared || p.group==ToolGroup) break;
	}
	if (i==Project._all.length)  return 0;
	else return p;
};

Project.merge = function(intoproject,otherproject) {
	// Save a copy of the file that is merged
	var savedcopy = exportProject(otherproject.id);
	var saveddate = otherproject.date;
	// Move each of the services over, one by one
	for (var i=0; i<otherproject.services.length; i++) {
		var s = Service.get(otherproject.services[i]);
		// Change the service, and create its DOM elements.
		s.setproject(intoproject.id);
		s.settitle(s.title);
		s.load();
		intoproject.services.push(s.id);

		var it = new ComponentIterator({service: s.id});
		for (it.first(); it.notlast(); it.next()) {
			var cm = it.getcomponent();
			cm.setproject(intoproject.id);
			var it2 = new ComponentIterator({project: intoproject.id, type: cm.type});
			for (it2.first(); it2.notlast(); it2.next()) {
				var cm2 = it2.getcomponent();
				if (cm2.id==cm.id) continue;
				if (!isSameString(cm2.title,cm.title)) continue;

				// There is already a component in this project with title cm.title
				// cm needs to be merged into cm2.
				cm2.absorbe(cm);
				break;	// Can occur only once, as cm2.title must be unique in otherproject
			}
		}
		
		// All nodes in this service need to be added to the proper NodeCluster in intoproject
		it = new NodeIterator({service: s.id});
		for (it.first(); it.notlast(); it.next()) {
			var rn = it.getnode();
			if (rn.type=='tACT' || rn.type=='tNOT') continue;

			cm = Component.get(rn.component);
			// If the node was added to a singular component, it doesn't need to added,
			// unless it is the first node in the singular class.
			if (cm.single && cm.nodes[0]!=rn.id) continue;

			for (var j=0; j<cm.thrass.length; j++) {
				var ta = ThreatAssessment.get(cm.thrass[j]);
				// During absorbe() in the loop above rn may already have been added to the
				// node clusters. Therefore, duplicateok==true in this call.
				NodeCluster.addnode_threat(intoproject.id,rn.id,ta.title,ta.type,true);
			}
		}
	}

	intoproject.store();
	otherproject.services = [];
	otherproject.destroy();
	// Now resurrect the other project
	i = loadFromString(savedcopy,true,false,'Merge');
	if (i==null) {
		bugreport("Failed to resurrect merged project", "Project.Merge");
	}
	otherproject = Project.get(i);
	otherproject.date = saveddate;

	if (intoproject.shared) {
		intoproject.storeOnServer(false,exportProject(intoproject.id),{});
	}
};






























































































































































Project.prototype = {
	destroy: function() {
		for (var i=0; i<this.threats.length; i++) {
			Threat.get(this.threats[i]).destroy();
		}
		for (i=0; i<this.services.length; i++) {
			Service.get(this.services[i]).destroy();
		}
		var it = new NodeClusterIterator({project: this.id});
		for (it.first(); it.notlast(); it.next()) {
			it.getNodeCluster().destroy();
		}
		localStorage.removeItem(LS+'P:'+this.id);
		Project._all[this.id]=null;
	},
	
	totalnodes: function() {
		var total=0;
		var it = new NodeIterator({project: this.id});
		for (it.first(); it.notlast(); it.next()) total++;
		return total;
	},
	
	settitle: function(newtitle) {
		newtitle = trimwhitespace(String(newtitle)).substr(0,50);
		if (newtitle==this.title)  return;
		if (newtitle=='')  return;
		var targettitle = newtitle;
		if (Project.withTitle(targettitle)!=null) {
			var n=0;
			do {
				targettitle = newtitle + " (" + (++n) + ")";
			} while (Project.withTitle(targettitle)!=null);
		}
		this.title = targettitle;
		this.store();
	},

	autosettitle: function() {
		this.settitle(_("New project"));
	},






















	setdescription: function(s) {
		this.description = trimwhitespace(String(s)).substr(0,100);
		this.store();
	},
	
	setdate: function(d) {
		this.date = trimwhitespace(String(d)).substr(0,20);
		this.store();
	},
	
	addservice: function(id) {
		var s = Service.get(id);
		if (this.services.indexOf(s.id)!=-1) {
			bugreport("service already added","Project.addservice");
		}
		s.setproject(this.id);
		this.services.push(s.id);
		this.store();
	},
	
	removeservice: function(id) {
		var s = Service.get(id);
		if (this.services.indexOf(s.id)==-1) {
			bugreport("no such service","Project.removeservice");
		}
		this.services.splice( this.services.indexOf(s.id),1 );
		s.destroy();
		this.store();
	},

	addthreat: function(id) {
		var th = Threat.get(id);
		if (this.threats.indexOf(th.id)!=-1) {
			bugreport("threat already added","Project.addthreat");
		}
		th.setproject(this.id);
		this.threats.push(th.id);
		this.store();
	},
	
	removethreat: function(id) {
		var th = Threat.get(id);
		if (this.threats.indexOf(th.id)==-1) {
			bugreport("no such threat","Project.removethreat");
		}
		this.threats.splice( this.threats.indexOf(th.id),1 );
		th.destroy();
		this.store();
	},
	
	unload: function() {
		for (var i=0; i<this.services.length; i++) {
			Service.get( this.services[i] ).unload();
		}
		$('#tWLSthreats').empty();
		$('#tWRDthreats').empty();
		$('#tEQTthreats').empty();
		$('.projectname').empty();
	},

	load: function() {
		for (var i=0; i<this.services.length; i++) {
			var s = Service.get( this.services[i] );
			s.load();
		}
		for (i=0; i<this.threats.length; i++) {
			var th = Threat.get( this.threats[i] );
			switch (th.type) {
			case 'tWLS':
				th.addtablerow('#tWLSthreats');
				break;
			case 'tWRD':
				th.addtablerow('#tWRDthreats');
				break;
			case 'tEQT':
				th.addtablerow('#tEQTthreats');
				break;
			default:
				bugreport('unknown type encountered','Project.load'); 
			}
		}

		var pid = this.id;
		var sortfunc = function(/*event,ui*/) {
			var p = Project.get(pid);
			var newlist = [];
			$('#tWLSthreats .threat').each( function(index,elem) {
				newlist.push( nid2id(elem.id) );
			});
			$('#tWRDthreats .threat').each( function(index,elem) {
				newlist.push( nid2id(elem.id) );
			});
			$('#tEQTthreats .threat').each( function(index,elem) {
				newlist.push( nid2id(elem.id) );
			});
			if (newlist.length != p.threats.length) {
				bugreport("internal error in sorting default vulnerabilities","Project.load");
			}
			p.threats = newlist;
			p.store();
			transactionCompleted("Project threats reordered "+pid);
		};
		$('#tWLSthreats').sortable({
			containment: 'parent',
			helper: 'clone',
			cursor: 'ns-resize',
			deactivate: sortfunc
		});
		$('#tWRDthreats').sortable({
			containment: 'parent',
			helper: 'clone',
			cursor: 'ns-resize',
			deactivate: sortfunc
		});
		$('#tEQTthreats').sortable({
			containment: 'parent',
			helper: 'clone',
			cursor: 'ns-resize',
			deactivate: sortfunc
		});

		$('.projectname').text(this.title);



		Project.cid = this.id;
		Service.cid = this.services[0];
// Do not paint, because we may want to show a tab other than the diagram
//		Service.get(Service.cid).paintall();
		Preferences.setcurrentproject(this.title);
	},
	
	adddefaultthreats: function() {
		for (var i=0; i<DefaultThreats.length; i++) {
			var th = new Threat(DefaultThreats[i][0]);
			th.setproject(this.id);
			th.settitle(DefaultThreats[i][1]);
			th.setdescription(DefaultThreats[i][2]);
			this.addthreat(th.id);
		}
	},


	strToLabel: function(str) {
		var i = Project.colors.indexOf(str);
		if (i==-1) {
			bugreport("Invalid color code","Project.strToLabel;");
			return "";
		}
		if (i==0)  return "";
		return this.labels[i-1];
	},
	
	_stringify: function() {
		var data = {};
		data.l=this.title;
		data.g=this.group;
		data.a=this.shared;
		data.s=this.services;
		data.d=this.description;
		data.w=this.date;
		data.t=this.threats;
		data.c=this.labels;
		return JSON.stringify(data);
	},

	exportstring: function() {
		var key = LS+'P:'+this.id;
		return key + '\t' + this._stringify() + '\n';
	},
	
	store: function() {
		if (this.stub)  return;
		var key = LS+'P:'+this.id;
		localStorage[key] = this._stringify();
	},





































































































































































































































































	internalCheck: function() {
		var errors = "";
		var it;
		if (this.stub && this.services.length>0) {
			errors += "Project is marked as a stub, but does have services.\n";
		}
		if (!this.stub && this.services.length==0) {
			errors += "Project is has no services.\n";
		}
		// Check each service, and all nodes in each service
		for (var i=0; i<this.services.length; i++) {
			var s = Service.get(this.services[i]);
			if (!s) {
				errors += "Service "+this.services[i]+" does not exist.\n";
				continue;
			}
			if (s.project != this.id) {
				errors += "Service "+s.id+" belongs to a different project.\n";
			}
			// Check all nodes for this service
			it = new NodeIterator({service: s.id});
			for (it.first(); it.notlast(); it.next()) {
				var rn = it.getnode();
				errors += rn.internalCheck();
			}
		}
		// Check all services that claim to belong to this project
		it = new ServiceIterator({project: this.id});
		for (it.first(); it.notlast(); it.next()) {
			s = it.getservice();
			if (this.services.indexOf(s.id)==-1) {
				errors += "Service "+s.id+" claims to belong to project "+this.id+" but is not known as a member.\n";
			}
		}
		// Check all node clusters
		it = new NodeClusterIterator({project: this.id});
		if (this.stub && it.notlast()) {
			errors += "Project is marked as a stub, but does have node clusters.\n";
		}
		for (it.first(); it.notlast(); it.next()) {
			var nc = it.getNodeCluster();
			if (!nc) {
				errors += "Node cluster "+it.getNodeClusterid()+" does not exist.\n";
				continue;
			}
			errors += nc.internalCheck();
		}
		// Check all components
		it = new ComponentIterator({project: this.id});
		if (this.stub && it.notlast()) {
			errors += "Project is marked as a stub, but does have components.\n";
		}
		for (it.first(); it.notlast(); it.next()) {
			var cm = it.getcomponent();
			if (!cm) {
				errors += "Component "+it.getcomponentid()+" does not exist.\n";
				continue;
			}
			errors += cm.internalCheck();
		}
		// Check all threats on this project
		if (this.stub && this.threats.length>0) {
			errors += "Project is marked as a stub, but does have a threats array.\n";
		}
		for (i=0; i<this.threats.length; i++) {
			var t = Threat.get(this.threats[i]);
			if (!t) {
				errors += "Threat "+this.threats[i]+" does not exist.\n";
				continue;
			}
			if (t.project != this.id) {
				errors += "Threat "+t.id+" belongs to a different project.\n";
			}
		}
		// Check all threats
		it = new ThreatIterator(this.id,'tUNK');
		if (this.stub && it.notlast()) {
			errors += "Project is marked as a stub, but does have threats.\n";
		}
		for (it.first(); it.notlast(); it.next()) {
			t = it.getthreat();
			if (!t) {
				errors += "Component "+it.getthreatid()+" does not exist.\n";
				continue;
			}
		}
		return errors;
	}
};

function escapeNewlines(s) {
	// Change backslash to "backslash-!" and newlines to "backslash-n"
	return s.replace(/\\/g,'\\!').replace(/\n/g,'\\n');
}

function unescapeNewlines(s) {
	return s.replace(/\\n/g,'\n').replace(/\\!/g,'\\');
}















































/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * ProjectIterator: iterate over all projects
 *
 * usage:
 * 		var it = new ProjectIterator({shared: true});
 *		it.sortByTitle();
 * 		for (it.first(); it.notlast(); it.next()) {
 *			var p = it.getproject();
 *	 		:
 *		}
 * Options:
 *	title: title matches string
 *	group: group matches string
 *	shared: only projects with this sharing status. Default: both shared and non-shared.
 *	stubsonly: true: only stubs, false: all projects. Undefined: only non-stubs.
 */
var ProjectIterator = function(opt) {
	this.index = 0;
	this.item = [];
	for (var i=0; i<Project._all.length; i++) {
		if (Project._all[i]==null) continue;
		var p =  Project._all[i];
		var ok = true;
		if (opt && opt.title!=null) {
			ok = ok && (isSameString(p.title,opt.title));
		}
		if (opt && opt.group!=null) {
			ok = ok && (!p.shared || p.group===opt.group);
		}
		if (opt && opt.shared!=null) {
			ok = ok && (p.shared===opt.shared);
		}
		if (opt && opt.stubsonly==null) {
			ok = ok && (p.stub==false);
		} else {
			ok = ok && (p.stub || !(opt && opt.stubsonly));
		}
		if (ok) {
			this.item.push(i);
		}
	}
};
ProjectIterator.prototype = {
	first: function() {this.index=0;},
	next: function() {this.index++;},
	number: function() {return this.item.length;},
	notlast: function() {return (this.index < this.item.length);},
	getprojectid: function() {return this.item[this.index];},
	getproject: function() {return Project.get( this.item[this.index] );},
	sortByTitle: function() {
		this.item.sort( function(a,b) {
			var pa = Project.get(a);
			var pb = Project.get(b);
			return pa.title.toLocaleLowerCase().localeCompare(pb.title.toLocaleLowerCase());
		});
	}
};

